#ifndef _FIRELINK_INTERFACE_H
#define _FIRELINK_INTERFACE_H

#include <stdint.h>
#include <exception>

#include <UsbInterface.h>



//exception that is thrown when an error exists in the Firelink interface
class FirelinkException: public std::exception
{
public:
	FirelinkException(const char* message) noexcept
	{
		mMessage = new char[strlen(message) + 1];
		strcpy(mMessage, message);
	}
	
	~FirelinkException() noexcept
	{
		delete[] mMessage;
	}

	const char* what() const noexcept
	{
		return mMessage;
	}

private:
	char* mMessage;
};



class FirelinkInterface: public virtual UsbInterface
{
public:
	enum SwdStatus: uint8_t
	{
		SWD_STATUS_SUCCESS = 0,
		SWD_STATUS_ACK_WAIT,
		SWD_STATUS_ACK_FAULT,
		SWD_STATUS_ACK_NONE,
		SWD_STATUS_PARITY_ERROR,
		SWD_STATUS_ERROR,
		SWD_STATUS_TIMEOUT,
	};
	
	enum ConnectorTG0Mux: uint8_t
	{
		FIRELINK_TG0_TRST = 0,
		FIRELINK_TG0_U2TX = 1
	};

	enum ConnectorTG1Mux: uint8_t
	{
		FIRELINK_TG1_DISABLED = 0,
		FIRELINK_TG1_U1TX = 1,
		FIRELINK_TG1_TDI = 2,
		FIRELINK_TG1_LOW = 3,
	};

	enum ConnectorTG3Mux: uint8_t
	{
		FIRELINK_TG3_DISABLED = 0,
		FIRELINK_TG3_SRST = 1,
		FIRELINK_TG3_TDI = 2,
		FIRELINK_TG3_U1TX = 3
	};

	enum ConnectorTG5Mux: uint8_t
	{
		FIRELINK_TG5_DISABLED = 0,
		FIRELINK_TG5_SRST = 1,
		FIRELINK_TG5_TDI = 2,
		FIRELINK_TG5_U1TX = 3
	};

	enum ConnectorTG7Mux: uint8_t
	{
		FIRELINK_TG7_TRST = 0,
		FIRELINK_TG7_U2TX = 1,
	};

	enum ConnectorTG8Mux: uint8_t
	{
		FIRELINK_TG8_DISABLED = 0,
		FIRELINK_TG8_U1TX = 1,
		FIRELINK_TG8_TDI = 2,
		FIRELINK_TG8_LOW = 3,
	};
	
	enum ConnectorU1RXMux: uint8_t
	{
		FIRELINK_U1RX_TG4 = 0,
		FIRELINK_U1RX_TDO = 1,
		FIRELINK_U1RX_TG9 = 2,
		FIRELINK_U1RX_TG11 = 3,
	};

	enum ConnectorU2RXMux: uint8_t
	{
		FIRELINK_U2RX_TG2 = 0,
		FIRELINK_U2RX_TG6 = 1,
		FIRELINK_U2RX_TG10 = 2
	};
	
	
	enum JtagTapState: uint8_t
	{
		JTAG_TAPS_RESET = 0x0,
		JTAG_TAPS_IDLE = 0x1,
		JTAG_TAPS_SELECT_DR = 0x2,
		JTAG_TAPS_SELECT_IR = 0x3,
		JTAG_TAPS_CAPTURE_DR = 0x4,
		JTAG_TAPS_CAPTURE_IR = 0x5,
		JTAG_TAPS_SHIFT_DR = 0x6,
		JTAG_TAPS_SHIFT_IR = 0x7,
		JTAG_TAPS_EXIT1_DR = 0x8,
		JTAG_TAPS_EXIT1_IR = 0x9,
		JTAG_TAPS_PAUSE_DR = 0xA,
		JTAG_TAPS_PAUSE_IR = 0xB,
		JTAG_TAPS_EXIT2_DR = 0xC,
		JTAG_TAPS_EXIT2_IR = 0xD,
		JTAG_TAPS_UPDATE_DR = 0xE,
		JTAG_TAPS_UPDATE_IR = 0xF,
		JTAG_TAPS_INVALID = 0x10
	};
	
	enum XsvfStatus: uint8_t
	{
		XSVF_STATUS_SUCCESS = 			0,				//everything is fine, hit the XCOMPLETE command (stream ended)
		XSVF_STATUS_PENDING = 			1,				//end of data encountered but not the end of the stream. Needs more data
		XSVF_STATUS_STREAM_ERROR = 	2,				//error in the XSVF stream
		XSVF_STATUS_MEM_ERROR = 		3,				//not enough memory reserved for bitstream encountered
		XSVF_STATUS_TDO_ERROR = 		4,				//TDO does not match expected TDO
		XSVF_STATUS_INVALID_ARG = 		5,
	};

	enum DebugPortRegisters: uint8_t
	{
		SWD_DP_ABORT		= 0x0,
		SWD_DP_DPIDR		= 0x0,
		SWD_DP_CSR			= 0x4,
		SWD_DP_RESEND		= 0x8,
		SWD_DP_SELECT		= 0x8,
		SWD_DP_RDBUFF		= 0xC,
		SWD_DP_TARGETSEL 	= 0xC,
		SWD_DP_DLCR			= 0x14,
		SWD_DP_TARGETID		= 0x24,
		SWD_DP_DLPIDR		= 0x34,
		SWD_DP_EVENTSTAT	= 0x44
	};

	enum DebugPortAbortRegister: uint8_t
	{
		//ABORT register definitions
		SWD_DP_ABORT_ORUNERRCLR		= (1 << 4),
		SWD_DP_ABORT_WDERRCLR		= (1 << 3),
		SWD_DP_ABORT_STKERRCLR		= (1 << 2),
		SWD_DP_ABORT_STKCMPCLR		= (1 << 1),
		SWD_DP_ABORT_DAPABORT		= (1 << 0),
		SWD_DP_ABORT_ALL			= 0x1F
	};

	enum DebugPortCsrRegister: uint32_t
	{
		//CSR register definitions
		SWD_DP_CSR_CSYSPWRUPACK		= (1U << 31),
		SWD_DP_CSR_CSYSPWRUPREQ		= (1 << 30),
		SWD_DP_CSR_CDBGPWRUPACK		= (1 << 29),
		SWD_DP_CSR_CDBGPWRUPREQ		= (1 << 28),
		SWD_DP_CSR_CDBGRSTACK		= (1 << 27),
		SWD_DP_CSR_CDBGRSTREQ		= (1 << 26),
		SWD_DP_CSR_TRNCNT_Pos		= 12,
		SWD_DP_CSR_TRNCNT_Msk		= (0xFFF << SWD_DP_CSR_TRNCNT_Pos),
		SWD_DP_CSR_MASKLANE_Pos		= 8,
		SWD_DP_CSR_MASKLANE_Msk		= (0xF << SWD_DP_CSR_MASKLANE_Pos),
		SWD_DP_CSR_WDATAERR			= (1 << 7),
		SWD_DP_CSR_READOK			= (1 << 6),
		SWD_DP_CSR_STICKYERR		= (1 << 5),
		SWD_DP_CSR_STICKYCMP		= (1 << 4),
		SWD_DP_CSR_TRNMODE_Pos		= 2,
		SWD_DP_CSR_TRNMODE_Msk		= (0x3 << SWD_DP_CSR_TRNMODE_Pos),
		SWD_DP_CSR_STICKYORUN		= (1 << 1),
		SWD_DP_CSR_ORUNDETECT		= (1 << 0),

		SWD_DP_CSR_XFER_NORMAL		= 0x00,
		SWD_DP_CSR_XFER_PUSHV		= 0x01,
		SWD_DP_CSR_XFER_PUSHCMP		= 0x02
	};

	enum AccessPortRegisters: uint8_t
	{
		SWD_AP_IDR				= 0xFC,
	};

	enum AccessPortIDR: uint32_t
	{
		SWD_AP_IDR_CLASS_Pos		= 13,
		SWD_AP_IDR_CLASS_Msk		= (0xF << SWD_AP_IDR_CLASS_Pos),
		
		SWD_AP_IDR_CLASS_MEMAP		= 0x8
	};

	enum MemAccessPortMapRegister: uint8_t
	{
		SWD_MEMAP_CSW		= 0x00,
		SWD_MEMAP_TAR		= 0x04,
		SWD_MEMAP_DRW		= 0x0C,
		SWD_MEMAP_BD0		= 0x10,
		SWD_MEMAP_BD1		= 0x14,
		SWD_MEMAP_BD2		= 0x18,
		SWD_MEMAP_BD3		= 0x1C,
		SWD_MEMAP_CFG		= 0xF4,
		SWD_MEMAP_BASE		= 0xF8
	};


	enum MemAccessPortCswRegister: uint8_t
	{
		SWD_MEMAP_CSW_ACCESS			= 0x07,
		SWD_MEMAP_CSW_ACCESS_BYTE		= 0x00,
		SWD_MEMAP_CSW_ACCESS_HALF		= 0x01,
		SWD_MEMAP_CSW_ACCESS_WORD		= 0x02,
		SWD_MEMAP_CSW_AIM				= 0x30,
		SWD_MEMAP_CSW_AIL_SCS_OFF		= 0x00,
		SWD_MEMAP_CSW_AIL_SCS_SINGLE	= 0x10,
		SWD_MEMAP_CSW_AIL_SCS_PACKED	= 0x20
	};

	enum UartStopBits: uint32_t
	{
		UART_SB_ONE = 		(0 << 12),
		UART_SB_HALF = 		(1 << 12),
		UART_SB_TWO = 		(2 << 12),
		UART_SB_ONEHALF = 	(3 << 12),
	};

	enum UartParity: uint32_t
	{
		UART_PARITY_NONE =	(0 << 9),
		UART_PARITY_EVEN =	(2 << 9),
		UART_PARITY_ODD =	(3 << 9),
		UART_PARITY_MARK =	(0 << 9),
		UART_PARITY_SPACE =	(0 << 9),
	};

	typedef void (*uart_receive_callback_t)(const uint8_t*, size_t);

#if defined (__GNUC__)
	//__attribute__ ((visibility("default")))
#elif defined (_MSC_VER)
#ifdef _BUILD_DLL
	__declspec(dllexport)
#else
	__declspec(dllimport)
#endif
#endif
	void check_swd_status(SwdStatus status);

#if defined (__GNUC__)
	//__attribute__ ((visibility("default")))
#elif defined (_MSC_VER)
#ifdef _BUILD_DLL
	__declspec(dllexport)
#else
	__declspec(dllimport)
#endif
#endif
	void check_xsvf_status(XsvfStatus status);

	typedef void (*event_callback_t)(uint16_t event, const void* payload, size_t size);
	

#if defined (__GNUC__)
//__attribute__ ((visibility("default")))
#elif defined (_MSC_VER)
#ifdef _BUILD_DLL
	__declspec(dllexport)
#else
	__declspec(dllimport)
#endif
#endif
	static FirelinkInterface* createInstance(void);

#if defined (__GNUC__)
	//__attribute__ ((visibility("default")))
#elif defined (_MSC_VER)
#ifdef _BUILD_DLL
	__declspec(dllexport)
#else
	__declspec(dllimport)
#endif
#endif
	virtual ~FirelinkInterface();


#if defined (__GNUC__)
	//__attribute__ ((visibility("default")))
#elif defined (_MSC_VER)
#ifdef _BUILD_DLL
	__declspec(dllexport)
#else
	__declspec(dllimport)
#endif
#endif
	virtual uint16_t getVersion(void) const;

	virtual void uart1Open(void) = 0;
	virtual void uart1Close(void) = 0;
	virtual void uart1Send(void* data, size_t size, uint32_t timeout) = 0;
	virtual int32_t uart1Recv(void* data, size_t size, uint32_t timeout) = 0;
	virtual void uart1Flush(void) = 0;

	virtual void uart2Open(void) = 0;
	virtual void uart2Close(void) = 0;
	virtual void uart2Send(void* data, size_t size, uint32_t timeout) = 0;
	virtual int32_t uart2Recv(void* data, size_t size, uint32_t timeout) = 0;
	virtual void uart2Flush(void) = 0;

	virtual void setEventCallback(event_callback_t event) = 0;

	virtual void resetProtocol(void) = 0;
	virtual void nop(void) = 0;
	virtual void delay_us(uint32_t us) = 0;
	
	virtual void updateFirmware(const void* image, size_t size) = 0;
	
	virtual void sequenceClear(void) = 0;
	virtual void sequenceRun(uint32_t timeout) = 0;
	
	virtual void connectorTG0(ConnectorTG0Mux mux) = 0;
	virtual void connectorTG1(ConnectorTG1Mux mux) = 0;
	virtual void connectorTG3(ConnectorTG3Mux mux) = 0;
	virtual void connectorTG5(ConnectorTG5Mux mux) = 0;
	virtual void connectorTG7(ConnectorTG7Mux mux) = 0;
	virtual void connectorTG8(ConnectorTG8Mux mux) = 0;
	virtual void connectorU1RX(ConnectorU1RXMux mux) = 0;
	virtual void connectorU2RX(ConnectorU2RXMux mux) = 0;

	virtual void set_speed(uint32_t speed, uint32_t &actual_speed) = 0;
	virtual void get_speed(uint32_t &speed) = 0;
	virtual void srst(bool state) = 0;

	virtual void vtgGetActualVoltage(uint32_t &voltage) = 0;
	virtual void vtgGetVoltage(uint32_t &voltage) = 0;
	virtual void vtgSetVoltage(uint32_t voltage, uint32_t &actual_voltage) = 0;
	virtual void vtgEnable(bool enable) = 0;

	virtual void vpwrEnable(bool enable) = 0;
	virtual void vpwrGetActualVoltage(uint32_t &voltage) = 0;

	virtual void configureUart1(uint32_t baudrate, UartStopBits stop_bits, UartParity parity, uint32_t word_length) = 0;
	virtual void configureUart2(uint32_t baudrate, UartStopBits stop_bits, UartParity parity, uint32_t word_length) = 0;

	//low level SWD functions
	virtual void swd_init(void) = 0;
	virtual void swd_clock(size_t cycles, bool swdio_state) = 0;
	virtual void swd_clock_bits(uint32_t count, uint64_t bits) = 0;

	virtual void swd_dp_read(uint32_t reg, uint32_t &value, SwdStatus* status = nullptr) = 0;
	virtual void swd_dp_write(uint32_t reg, uint32_t value, SwdStatus* status = nullptr) = 0;

	virtual void swd_ap_read(uint32_t ap, uint32_t reg, uint32_t &value, SwdStatus* status = nullptr) = 0;
	virtual void swd_ap_write(uint32_t ap, uint32_t reg, uint32_t value, SwdStatus* status = nullptr) = 0;

	virtual void swd_connect(uint32_t &id, uint32_t timeout, SwdStatus* status = nullptr) = 0;

//MEM-AP functions
	virtual void swd_memap_read32(uint32_t ap, uint32_t address, uint32_t &value, SwdStatus* status = nullptr) = 0;
	virtual void swd_memap_write32(uint32_t ap, uint32_t address, uint32_t value, SwdStatus* status = nullptr) = 0;
	
	virtual void swd_memap_read16(uint32_t ap, uint32_t address, uint16_t &value, SwdStatus* status = nullptr) = 0;
	virtual void swd_memap_write16(uint32_t ap, uint32_t address, uint16_t value, SwdStatus* status = nullptr) = 0;
	
	virtual void swd_memap_read8(uint32_t ap, uint32_t address, uint8_t &value, SwdStatus* status = nullptr) = 0;
	virtual void swd_memap_write8(uint32_t ap, uint32_t address, uint8_t value, SwdStatus* status = nullptr) = 0;

	virtual void swd_memap_read_block(uint32_t ap, uint32_t address, uint32_t* data, uint32_t size, SwdStatus* status = nullptr) = 0;
	virtual void swd_memap_write_block(uint32_t ap, uint32_t address, const uint32_t* data, uint32_t size, SwdStatus* status = nullptr) = 0;

//SCS functions
	virtual void cortex_scs_debug_start(uint32_t mem_ap, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_debug_disable(uint32_t mem_ap, SwdStatus* status = nullptr) = 0;

	virtual void cortex_scs_read_register(uint32_t mem_ap, uint32_t reg, uint32_t &value, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_read_register_bank(uint32_t mem_ap, uint32_t* regs, size_t count, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_write_register(uint32_t mem_ap, uint32_t reg, uint32_t value, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_write_register_bank(uint32_t mem_ap, const uint32_t* regs, size_t count, SwdStatus* status = nullptr) = 0;

	virtual void cortex_scs_request_halt(uint32_t mem_ap, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_is_halted(uint32_t mem_ap, bool &halted, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_wait_halted(uint32_t mem_ap, uint32_t &pc, uint32_t timeout, SwdStatus* status = nullptr) = 0;

	virtual void cortex_scs_continue(uint32_t mem_ap, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_continue_at(uint32_t mem_ap, uint32_t address, SwdStatus* status = nullptr) = 0;	

	virtual void cortex_scs_step(uint32_t mem_ap, bool maskints, SwdStatus* status = nullptr) = 0;
	virtual void cortex_scs_step_at(uint32_t mem_ap, uint32_t address, bool maskints, SwdStatus* status = nullptr) = 0;	

	virtual void cortex_scs_request_system_reset(uint32_t mem_ap, SwdStatus* status = nullptr) = 0;


	//SPI functions
	virtual void spi_init(bool cpol, bool cpha, bool lsb_first, uint8_t width) = 0;
	virtual void spi_ss0(bool state) = 0;
	virtual void spi_ss1(bool state) = 0;

	virtual void spi_shift(const void* data_mosi, void* data_miso, size_t size) = 0;	
	virtual void spi_shift_byte(uint8_t mosi, uint8_t* miso) = 0;
	
	virtual void spif_wait_busy(uint8_t command, uint8_t mask, uint8_t &status_reg, uint32_t timeout) = 0;
	
	//JTAG functions
	virtual void jtag_init(void) = 0;
	virtual void jtag_clock(size_t cycles) = 0;
	virtual void jtag_trst(bool state) = 0;
	virtual void jtag_tms(size_t count, uint64_t value) = 0;
	virtual void jtag_state(JtagTapState state) = 0;
	
	virtual void jtag_shift(const void* tdi, void* tdo, size_t count) = 0;
	virtual void jtag_shift_multiple(const void* tdi, void* tdo, size_t count, size_t width) = 0;
	
	virtual void jtag_play_xsvf(const void* xsvf_image, size_t size, XsvfStatus &status) = 0;
};

#endif

